#   pyRCV2: Preferential vote counting
#   Copyright © 2020–2021  Lee Yingtong Li (RunasSudo)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pytest_steps import test_steps

import pyRCV2.blt
import pyRCV2.numbers
from pyRCV2.method.base_stv import EGSTVCounter
from pyRCV2.model import CandidateState, CountCompleted

import csv

# Read model CSV

with open('tests/data/ers97.csv', 'r', newline='') as f:
	reader = csv.reader(f)
	data = [x for x in reader]

candidates = [data[i][0] for i in range(2, len(data) - 2)]

@test_steps(*['Stage {}'.format(data[0][i]) for i in range(1, len(data[0]), 2)])
def test_aec_tas19():
	"""Compare count of ers97.blt with model result computed by hand"""
	
	pyRCV2.numbers.set_numclass(pyRCV2.numbers.Fixed)
	pyRCV2.numbers.set_dps(5)
	
	with open('tests/data/ers97.blt', 'r') as f:
		election = pyRCV2.blt.readBLT(f.read())
	
	assert len(election.candidates) == len(candidates)
	
	counter = EGSTVCounter(election, {
		'quota': 'droop_exact',
		'quota_mode': 'ers97',
		'bulk_exclude': True,
		'defer_surpluses': True,
		'round_quota': 2,
		'round_votes': 2,
		'round_tvs': 2,
		'round_weights': 2,
		'papers': 'transferable',
		'exclusion': 'by_value',
	})
	result = counter.reset()
	assert result.quota == pyRCV2.numbers.Num('107.58')
	
	for i in range(1, len(data[0]), 2):
		stage = int(data[0][i])
		
		while len(counter.step_results) < stage:
			result = counter.step()
		
		comment = data[1][i]
		assert result.comment == comment, 'Failed to verify comment'
		
		for j, cand in enumerate(candidates):
			votes = pyRCV2.numbers.Num(data[j + 2][i])
			cc = next(cc for c, cc in result.candidates.items() if c.name == cand)
			assert cc.votes == votes, 'Failed to verify candidate "{}" votes, got {} expected {}'.format(cand, cc.votes.pp(0), votes.pp(0))
			
			state = data[j + 2][i + 1] if len(data[j + 2]) > (i + 1) else ''
			accept = {'': CandidateState.HOPEFUL, 'PEL': CandidateState.PROVISIONALLY_ELECTED, 'EL': CandidateState.ELECTED, 'EX': CandidateState.EXCLUDED, 'EXCLUDING': CandidateState.EXCLUDING}
			assert cc.state == accept[state], 'Failed to verify candidate "{}" state'.format(cand)
		
		nontransferable = pyRCV2.numbers.Num(data[len(candidates) + 2][i])
		assert (result.exhausted.votes + result.loss_fraction.votes) == nontransferable, 'Failed to verify non-transferable votes, got {} expected {}'.format((result.exhausted.votes + result.loss_fraction.votes).pp(2), nontransferable.pp(2))
		
		if data[len(candidates) + 3][i]:
			votes_required = pyRCV2.numbers.Num(data[len(candidates) + 3][i])
			assert result.vote_required_election.pp(2) == votes_required.pp(2), 'Failed to verify votes required for election, got {} expected {}'.format(result.vote_required_election.pp(2), votes_required.pp(2))
		
		yield 'Stage {}'.format(stage)
